/**
 * Invoke Shell Verb NSIS plug-in
 *
 * Version: 1.0
 * Date   : 2nd February 2011
 * Author : Robert Strong
 */

// Uncomment to force Unicode support
//#define UNICODE
//#define _UNICODE

#include <windows.h>
#include "..\ExDll\exdll.h"
#include <shldisp.h>

#define OUT_SUCCESS            TEXT("success")
#define OUT_ERR_CALL_FAILED    TEXT("method failed")
#define OUT_ERR_INVALID_RESID  TEXT("invalid resource")
#define OUT_ERR_INVALID_DIR    TEXT("invalid directory")
#define OUT_ERR_INVALID_FILE   TEXT("invalid file")
#define OUT_ERR_INVALID_PARAM  TEXT("invalid parameter")
#define OUT_ERR_VERB_NOT_FOUND TEXT("verb not found")

extern "C"
void __declspec(dllexport) DoIt(HWND hWndParent, int string_size, 
                                TCHAR *variables, stack_t **stacktop,
                                extra_parameters *extra)
{
  EXDLL_INIT();
  {
    TCHAR szDirPath[MAX_PATH];
    TCHAR szFileName[MAX_PATH];
    TCHAR szResID[10];

    if (popstring(szDirPath) == 0 && popstring(szFileName) == 0 &&
        popstring(szResID) == 0)
    {
      HRESULT hr;
      HINSTANCE hInst;
      WCHAR wszVerbName[100];
      VARIANT v;
      BSTR name;
      BOOL bFoundVerb = FALSE;
      int rv = 0;
      IShellDispatch *pSD = NULL;
      Folder *pFolder = NULL;
      FolderItem *pItem = NULL;
      FolderItemVerbs *pVerbs = NULL;
      FolderItemVerb *pVerb = NULL;

      // Get the localized verb name from shell32.dll
      hInst = LoadLibraryW(L"shell32.dll");
      rv = LoadStringW(hInst,
#ifdef _UNICODE
                       _wtoi(szResID),
#else
                       atoi(szResID),
#endif
                       wszVerbName,
                       sizeof(wszVerbName)/sizeof(wszVerbName[0])); 
      FreeLibrary(hInst);
      if (!rv)
      {
        pushstring(OUT_ERR_INVALID_RESID);
        return;
      }

      CoInitialize(NULL);

      hr = CoCreateInstance(CLSID_Shell,
                            NULL,
                            CLSCTX_INPROC_SERVER,
                            IID_IShellDispatch,
                            (void **)&pSD);
      if (FAILED(hr) || pSD ==  NULL)
      {
        pushstring(OUT_ERR_CALL_FAILED);
        CoFreeUnusedLibraries();
        CoUninitialize();
        return;
      }

      v.vt = VT_BSTR;
#ifdef _UNICODE
      v.bstrVal = szDirPath;
#else
      WCHAR wszDirPath[MAX_PATH];
      MultiByteToWideChar(CP_ACP, 0, szDirPath, lstrlen(szDirPath) + 1,
                          wszDirPath, sizeof(wszDirPath)/sizeof(wszDirPath[0]));
      v.bstrVal = wszDirPath;
#endif

      hr = pSD->NameSpace(v, &pFolder);
      if (FAILED(hr) || pFolder == NULL)
      {
        pushstring(OUT_ERR_INVALID_DIR);
        pSD->Release();
        CoFreeUnusedLibraries();
        CoUninitialize();
        return;
      }

#ifdef _UNICODE
      hr = pFolder->ParseName(szFileName, &pItem);
#else
      WCHAR wszFileName[MAX_PATH];
      MultiByteToWideChar(CP_ACP, 0, szFileName, lstrlen(szFileName) + 1,
                          wszFileName, sizeof(wszFileName)/sizeof(wszFileName[0]));
      hr = pFolder->ParseName(wszFileName, &pItem);
#endif
      if (FAILED(hr) || pItem == NULL)
      {
        pushstring(OUT_ERR_INVALID_FILE);
        pSD->Release();
        CoFreeUnusedLibraries();
        CoUninitialize();
        return;
      }

      hr = pItem->Verbs(&pVerbs);
      if (FAILED(hr) || pVerbs == NULL)
      {
        pushstring(OUT_ERR_CALL_FAILED);
        pItem->Release();
        pSD->Release();
        CoFreeUnusedLibraries();
        CoUninitialize();
        return;
      }

      long cnt;
      hr = pVerbs->get_Count(&cnt);
      if (FAILED(hr))
      {
        pushstring(OUT_ERR_CALL_FAILED);
        pItem->Release();
        pSD->Release();
        CoFreeUnusedLibraries();
        CoUninitialize();
        return;
      }

      int i;
      v.vt = VT_I4;
      for (i = 0; i < cnt; ++i)
      {
        v.lVal = i;
        hr = pVerbs->Item(v, &pVerb);
        if (FAILED(hr) || pVerb == NULL)
        {
          continue;
        }

        hr = pVerb->get_Name(&name);
        if (FAILED(hr))
        {
          pVerb->Release();
          continue;
        }

        if (!wcscmp((WCHAR *)name, wszVerbName))
        {
          bFoundVerb = TRUE;
          hr = pVerb->DoIt();
          if (FAILED(hr))
          {
            pushstring(OUT_ERR_CALL_FAILED);
          }
        }

        SysFreeString(name);
        pVerb->Release();
        if (bFoundVerb)
        {
          break;
        }
      }

      if (bFoundVerb)
      {
        pushstring(OUT_SUCCESS);
      } else {
        pushstring(OUT_ERR_VERB_NOT_FOUND);
      }

      pVerbs->Release();
      pItem->Release();
      pFolder->Release();
      pSD->Release();

      CoFreeUnusedLibraries();
      CoUninitialize();
    }
  }
}

BOOL WINAPI DllMain(HANDLE hInstNew,
                    ULONG ul_reason_for_call,
                    LPVOID lpReserved)
{
    return TRUE;
}
